/*
* CommandInterpreter.java -  Provide the basic command line interface.
*
* Copyright (c) 1996 Chuck McManis, All Rights Reserved.
*
* Permission to use, copy, modify, and distribute this software
* and its documentation for NON-COMMERCIAL purposes and without
* fee is hereby granted provided that this copyright notice
* appears in all copies.
*
* CHUCK MCMANIS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
* SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. CHUCK MCMANIS
* SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT
* OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
*/

package basic;

import java.io.*;
import jp.gr.java_conf.sol.basic.*;
import jp.gr.java_conf.sol.basic.mz700.*;

/**
* This class is an "interactive" BASIC environment. You can think of it as
* BASIC debug mode. Using the streams you passed in to create the object, it
* hosts an interactive session allowing the user to enter BASIC programs, run
* them, save them, and load them.
*/
public class CommandInterpreter extends Thread {

  private BasicInputStream inStream;
  private BasicPrintStream outStream;
  private Program pgm;
  private boolean shouldStop;

  char data[] = new char[256];


  public CommandInterpreter(Console console) {

    super();
    this.inStream = new BasicInputStream(console.getInputStream());
    this.outStream= new BasicPrintStream(console.getOutputStream(), console);
    console.setInterpreter(this);
    pgm = new Program();
  }

  public void breakExecution() {

    if (!pgm.isRunning()) {
      outStream.write("\n");
      return;
    }

    pgm.setShouldStop(true);

    // wait until program stop
    try {
      synchronized(this) {
        this.wait();
      }
    } catch (InterruptedException ex) {
      ex.printStackTrace();
    }

    Statement s = pgm.getBreakPoint();
    if (s != null) {
      int lineNo = s.lineNo();
      outStream.write("\nBREAK IN "+ lineNo+"\n");
    } else {
      outStream.write("\nBREAK\n");
    }
  }

  public void run() {

    LexicalTokenizer lt = new LexicalTokenizer(data);

    BasicInputStream dis = inStream;
    BasicString lineData;

    while (true) {
      Statement s = null;
      try {
        lineData = dis.readLine();

      } catch (IOException ioe) {
        outStream.write("Caught an IO exception reading the input stream!\n");
        return;
      }

      // exit on eof of the input stream
      if (lineData == null)
      return;

      // ignore blank lines.
      if (lineData.length() == 0)
      continue;

      lt.reset(lineData);

      if (! lt.hasMoreTokens())
      continue;

      Token t = lt.nextToken();
      switch (t.typeNum()) {

        /*
        * Process an initial number, it can be a new statement line
        * or it may be an implicit delete command.
        */
        case Token.CONSTANT :

        Token peek = lt.nextToken();

        if (peek.typeNum() == Token.EOL) {
          pgm.del((int) t.numValue());
          break;
        } else {
          lt.unGetToken();
        }
        try {
          s = ParseStatement.statement(lt);
          s.addText(lineData);
          s.addLine((int) t.numValue());
          pgm.add((int) t.numValue(), s);

        } catch (BASICSyntaxError e) {   // fail to parse

          s = ParseStatement.parseAsUnknwonStatement(lt);

          s.addText(lineData);
          s.addLine((int) t.numValue());
          pgm.add((int) t.numValue(), s);

          continue;
        }
        break;

        /*
        * If initially it is a variable or a statement keyword then it
        * must be an 'immediate' line.
        */
        case Token.VARIABLE :
        case Token.STATEMENT : // immediate mode
        case Token.SYMBOL :

        lt.unGetToken();

        try {
        
          pgm.setRunning(true);
          s = ParseStatement.statement(lt);
          do {
            s = s.doit(pgm, inStream, outStream);
          } while (!pgm.isShouldStop()  && s!= null);

          pgm.setRunning(false);

          // if suspended by user by pushing break or stop key.
          if (pgm.isShouldStop()) {

            // save current statement
            pgm.setBreakPoint(s);

            pgm.setShouldStop(false);
            synchronized(this) {
              this.notifyAll();
            }
          } else {  // program end
            pgm.setBreakPoint(null);
          }

          outStream.write("READY\n");

        } catch (BASICSyntaxError e) {
          e.printStackTrace();
          outStream.write("SYNTAX ERROR\n");
          outStream.write("READY\n");
          continue;
        } catch (BASICRuntimeError er) {
          outStream.write(er.getMsg()+"\n");
          outStream.write("READY\n");
        }

        break;
        /*
        * Blank lines are ignored.
        */
        case Token.EOL :
        break;

        /*
        * Anything else is an error.
        */
        default:
        outStream.write("READY\n");
        break;
      }
    }
  }


}
