/*
* Statement.java - BASIC Statement object
*
* Copyright (c) 1996 Chuck McManis, All Rights Reserved.
*
* Permission to use, copy, modify, and distribute this software
* and its documentation for NON-COMMERCIAL purposes and without
* fee is hereby granted provided that this copyright notice
* appears in all copies.
*
* CHUCK MCMANIS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
* SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. CHUCK MCMANIS
* SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT
* OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
*/
package basic;

import java.io.*;
import java.util.Vector;
import util.RedBlackTree;
import java.util.Enumeration;

import java.lang.reflect.*;
import jp.gr.java_conf.sol.basic.*;

/**
* This class defines BASIC statements. As with expressions, there is a
* subclass of Statement that does parsing called ParseStatement. This
* separation allows the statement object to be half as large as it might
* otherwise be.
*
* The <i>execute</i> interface defines what the BASIC statements *do*.
* These are all called by the containing <b>Program</b>.
*/
abstract class Statement implements StatementInterface {

  protected String keyword; // type of statement
  protected int line;
  Statement nxt;  // if there are chained statements

  private BasicString orig;    // original string that was parsed into this statement.

  private boolean connectbyColon; // While parsing source, whether this statement can be connected to next statement using colon(":")
  private boolean endsByNull;    // While parsing source, whether this statement can ends by null
  private boolean endsByEOL;     // while parsing source, whether this statement can ends by EOL

  /**
  * Construct a new statement object with a valid key.
  */
  protected Statement(String key, boolean connectByColon ,boolean endsByNull, boolean endsByEOL) {
    keyword = key;
    this.connectbyColon = connectByColon;
    this.endsByNull = endsByNull;
    this.endsByEOL = endsByEOL;
  }


  /**
  * Return a string representation of this statement. If the
  * original text was set then use that, otherwise reconstruct
  * the string from the parse tree.
  */
  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("BASIC Statement :");
    if (orig != null) {
      sb.append(orig);
    } else {
      sb.append(unparse());
    }
    return sb.toString();
  }

  /**
  * Put a reference to the original string from which this statement was parsed.
  * this can be used for listing out the program in the native case style of the user.
  */
  void addText(BasicString t) { orig = t; }

  /**
  * Return the statement as a string.
  */
  BasicString asString() { return orig; }

  public String getKeyword() {
    return keyword;
  }

  /**
  * Update line number information in this statement. Used to determine the next
  * line to execute.
  */
  void addLine(int l) {
    line = l;
    if (nxt != null)
    nxt.addLine(l);
  }

  /**
  * Return this statements line number.
  */
  int lineNo() { return line; }

  /**
  * reconstruct the statement from the parse tree, this is most useful for
  * diagnosing parsing problems.
  */
  public String unparse(){
    return "unparse";
  }

  /**
  * This method "runs" this statement and returns a reference on
  * the next statement to run or null if there is no next statement.
  *
  * @throws BASICRuntimeError if there is a problem during statement
  * execution such as divide by zero etc.
  */
  public abstract Statement doit(Program pgm, InputStream in, BasicPrintStream out) throws BASICRuntimeError;

  protected RedBlackTree vars;

  // variables used by this statement.
  /**
  * The trace method can be used during execution to print out what
  * the program is doing.
  */
  void trace(Program pgm, BasicPrintStream ps) {
    StringBuffer sb = new StringBuffer();
    String n;
    sb.append("**:");

    if (vars == null)
    vars = getVars();

    /*
    * Print the line we're executing on the output stream.
    */
    n = line+"";
    for (int zz = 0; zz < 5 - n.length(); zz++)
    sb.append(' ');
    sb.append(n);
    sb.append(':');
    sb.append(unparse());
    ps.write(sb.toString()+"\n");
    if (vars != null) {
      for (Enumeration e = vars.elements(); e.hasMoreElements(); ) {
        VariableExpression vi = (VariableExpression) e.nextElement();
        BasicString  t = null;
        try {
          t = vi.stringValue(pgm);
        } catch (BASICRuntimeError bse) {
          t = new BasicString("Not yet defined.");
        }
        ps.write("        :"+vi.unparse()+" = "+(vi.isString() ? "\""+t+"\"" : t.getAsciiString()));
        ps.write("\n");
      }
    }
  }

  /**
  * Can be overridden by statements that use variables in their execution.
  */
  RedBlackTree getVars() {
    return null;
  }

  public boolean isEndsByNull() {
    return endsByNull;
  }

  public boolean isEndsByEOL() {
    return endsByEOL;
  }

  public boolean isConnectbyColon() {
    return connectbyColon;
  }

  /** create Object of class basic.NAMEStatement using reflection
  */
  public static Statement createInstance(String name, LexicalTokenizer lt) {

    Statement s = null;

    try {
      Class clazz = Class.forName("basic."+name+"Statement");
      Class[] constructorType = new Class[1];
      constructorType[0] = Class.forName("basic.LexicalTokenizer");

      Constructor constructor = clazz.getConstructor(constructorType);
      Object[] arg = new Object[1];
      arg[0] = lt;
      
      s = (Statement)constructor.newInstance(arg);

    } catch (Exception ex) {
      ex.printStackTrace();
      if (ex instanceof InvocationTargetException) {
        InvocationTargetException ex2 = (InvocationTargetException)ex;
        Throwable ex3 = ex2.getTargetException();
        ex3.printStackTrace();
      }
      System.out.println("Fail to create "+name+"Statement");
      try {
        s = new UnknownStatement(name, lt);
      } catch (BASICSyntaxError ex2) {

      }
    }

    return s;
  }


  Token[] parseTokens(LexicalTokenizer lt) {

    Vector v = new Vector();

    boolean finish = false;
    while(!finish) {
      lt.mark();
      Token t = lt.nextToken();
      if (t == null) {
        //lt.resetToMark();
        finish = true;
      } else if (t.typeNum() == Token.EOL) {
        //lt.resetToMark();
        finish = true;
      } else if (t.typeNum() == Token.SYMBOL && t.numValue() == ':') {
        lt.resetToMark();
        finish = true;
      } else if (t.typeNum() == Token.ERROR) {
        finish = true;
      } else {
        v.addElement(t);
      }
    }

    if (v.size() == 0) {
      return null;
    }

    Token[] tokens = new Token[v.size()];
    v.copyInto(tokens);

    return tokens;
  }

  protected boolean isComma(Token t) {
    return (t.typeNum() == Token.SYMBOL && t.numValue() == 44);
  }
  protected boolean isValue(Token t) {
    return (t.typeNum() == Token.CONSTANT);
  }
}


